"""
Copied from https://github.com/pypiserver/pypiserver
Under MIT license

Removed unused code lines
"""

import re
from urllib.parse import quote


def normalize_pkgname(name: str) -> str:
    """Perform PEP 503 normalization"""
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_pkgname_for_url(name: str) -> str:
    """Perform PEP 503 normalization and ensure the value is safe for URLs."""
    return quote(normalize_pkgname(name))



from abc import abstractmethod, ABC
from typing import List, Optional

from warehouse14.models import Project, Account, Token, Group


class DBBackend(ABC):

    # Account methods
    @abstractmethod
    def account_save(self, user_id: str, **kwargs) -> Optional[Account]:
        """
        Load non confidential data for a user
        :param user_id: ID of the user, who is represented by a account
        :param kwargs: further properties of the user
        """

    @abstractmethod
    def account_get(self, user_id: str) -> Optional[Account]:
        """
        Load non confidation data for a user
        :param user_id: ID of the user, which will be used to search for
        """

    @abstractmethod
    def account_token_add(
        self, user_id: str, token_id: str, name: str, key: str
    ) -> Optional[Token]:
        """
        Add a new API Token to an account
        """

    @abstractmethod
    def account_token_list(self, account_id: str) -> List[Token]:
        """
        List all API tokens of account
        """

    @abstractmethod
    def account_token_delete(self, account_id: str, token_id: str):
        """
        Delete API token from account
        """

    @abstractmethod
    def resolve_token(self, token_id: str) -> Optional[Account]:
        """
        Look up a account by given token id.

        :param token_id: Unique token id
        :return: user_id
        """

    # Group methods
    @abstractmethod
    def group_create(self, name: str, admins: List[str]) -> Group:
        """
        Create a new group.

        :param name: Unique group name
        :param admins: List of admins of the group
        :return: Group
        """

    @abstractmethod
    def group_get(self, name: str) -> Optional[Group]:
        """
        Get group by name.

        :return: Group
        """

    @abstractmethod
    def account_groups_list(self, user_id: str) -> List[str]:
        """
        List all group names a user belongs to.

        :return: Names of Group
        """

    @abstractmethod
    def group_delete(self, name: str):
        """
        Deletes a group
        """

    # Project methods
    @abstractmethod
    def project_save(self, project: Project) -> Project:
        """
        Create a project in the package index
        :param project: project data
        """

    def project_get(self, name: str) -> Optional[Project]:
        """
        Returns project by given name.

        May be overwritten by subclasses, for an optimized implementation.
        :param name: project name
        :return: Project or None
        """

    @abstractmethod
    def project_list(self) -> List[Project]:
        """
        Lists all project names
        """



from flask import Flask, render_template
from flask_login import login_required, current_user

from warehouse14.repos import DBBackend


def get_user_id():
    return current_user.account.name


def add_routes(app: Flask, db: DBBackend):
    @app.route("/groups")
    @login_required
    def groups_list():
        user_id = get_user_id()

        groups = db.account_groups_list(user_id)
        return render_template("group/groups.html", groups=groups)



import datetime
import hashlib
import re
from typing import List, Optional, Dict, Literal
from urllib.parse import quote

from pydantic import BaseModel

PackageType = Literal[
    "bdist_dmg",
    "bdist_dumb",
    "bdist_egg",
    "bdist_msi",
    "bdist_rpm",
    "bdist_wheel",
    "bdist_wininst",
    "sdist",
]


#
class File(BaseModel):
    # key: str
    # python_version: str
    # requires_python: str
    # packagetype: PackageType
    # comment_text: str
    filename: str
    # size: int
    # md5_digest: str
    sha256_digest: str
    # blake2_256_digest: datetime.datetime
    # upload_time: str
    # uploaded_via: str


#
# class VersionV2(BaseModel):
#     version: str
#
#     is_prerelease: bool
#     author: str
#     author_email: str
#     maintainer: str
#     maintainer_email: str
#     home_page: str
#     license: str
#     summary: str
#     keywords: str
#     platform: str
#     download_url: str
#     requires_python: str
#     created: datetime.datetime
#     description: str
#     classifiers: List[str]
#     files: List[File]


class Version(BaseModel):
    version: str
    metadata: dict = {}
    files: List[File] = list()

    @property
    def author(self):
        return self.metadata.get("author", "")

    @property
    def maintainer(self):
        return self.metadata.get("maintainer", "")

    @property
    def summary(self):
        return self.metadata.get("summary", "")

    @property
    def description(self) -> str:
        return self.metadata.get("description", "")

    @property
    def description_content_type(self) -> str:
        return self.metadata.get("description_content_type", "")

    @property
    def classifiers(self) -> List[str]:
        return self.metadata.get("classifiers", [])


class Token(BaseModel):
    id: str
    name: str
    key: str
    created: datetime.datetime


class Account(BaseModel):
    name: str
    created: datetime.datetime
    updated: datetime.datetime

    @staticmethod
    def hash_name(name: str):
        return hashlib.sha256(name.encode()).hexdigest()


class Group(BaseModel):
    name: str
    # created: datetime.datetime
    # updated: datetime.datetime
    admins: List[str] = []


class Project(BaseModel):
    # Primary key
    def normalized_name(self) -> str:
        """Perform PEP 503 normalization"""
        return Project.normalize_name(self.name)

    name: str
    admins: List[str] = []
    members: List[str] = []
    public: bool = False
    versions: Dict[str, Version] = {}

    @property
    def latest_version(self) -> Optional[Version]:
        if not self.versions:
            return None

        version = sorted(self.versions.keys())
        return self.versions[version[-1]]

    @property
    def files(self):
        return [file for version in self.versions.values() for file in version.files]

    def add_file(self, version: str, file: File):
        self.versions.setdefault(version, Version(version=version)).files.append(file)

    def visible(self, user: str):
        return self.public or (user in self.admins + self.members)

    def is_admin(self, user: str):
        return user in self.admins

    def normalized_name_for_url(self) -> str:
        """Perform PEP 503 normalization and ensure the value is safe for URLs."""
        return quote(self.normalized_name())

    @staticmethod
    def normalize_name(name: str) -> str:
        """Perform PEP 503 normalization"""
        return re.sub(r"[-_.]+", "-", name).lower()


# class DB(ABC):
#     def get_project(self, name: str) -> Optional[Project]:
#         """
#         Returns project by given name.
#
#         May be overwritten by subclasses, for an optimized implementation.
#         :param name: project name
#         :return: Project or None
#         """
#         name = Project.normalize_name(name)
#         for project in self.projects():
#             if project.normalized_name() == name:
#                 return project
#         else:
#             return None
#
#     @abstractmethod
#     def projects(self) -> List[Project]:
#         """
#         Lists all project names
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def create_project(self, project: Project) -> Project:
#         """
#         Create a project in the package index
#         :param project: project data
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def update_project(self, project: Project) -> Project:
#         """
#         Update a project in the package index
#         :param project: project data
#         :raises Exception - in case project does not exists
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def delete_project(self, name: str):
#         """
#         Deletes a project from the package index
#         :param name: Name of the project
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def save_account(self, account: Account) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user: User object to be saved in db
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def load_account(self, user_id: str) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user_id: ID of the user, which will be used to search for
#         """
#         raise NotImplementedError()
#
#     @abstractmethod
#     def resolve_token(self, token_id: str) -> Optional[Account]:
#         """
#         Look up a account by given token id.
#
#         :param token_id: Unique token id
#         :return: user_id
#         """
#         raise NotImplementedError
#
#
# class SimpleFileDB(DB):
#     def __init__(self, root: Union[str, Path]):
#         super().__init__()
#         self._root = Path(root).expanduser().resolve()
#
#         self._project_root = self._root / "projects"
#         self._account_root = self._root / "accounts"
#
#     def _save_project(self, project: Project):
#         file = self._locate_project(project.name)
#         file.parent.mkdir(parents=True, exist_ok=True)
#
#         with file.open("wt") as f:
#             f.write(project.json())
#
#     def _locate_project(self, name: str) -> Path:
#         return self._project_root / f"{Project.normalize_name(name)}.json"
#
#     def _locate_account(self, name) -> Path:
#         return self._account_root / f"{Account.hash_name(name)}.json"
#
#     def _list_projects(self) -> List[Project]:
#         project_files = [x for x in self._project_root.glob("**/*") if x.is_file()]
#         return [self._load_project(file) for file in project_files]
#
#     def _map_token_to_account(self) -> Dict[str, Account]:
#         tokens = {}
#         account_files = [x for x in self._account_root.glob("**/*") if x.is_file()]
#         for file in account_files:
#             account = Account.parse_file(file)
#             for token in account.tokens:
#                 tokens[token.id] = account
#         return tokens
#
#     @staticmethod
#     def _load_project(file: Union[str, Path]) -> Project:
#         return Project.parse_file(Path(file))
#
#     def create_project(self, project: Project):
#         self._save_project(project)
#         return project
#
#     def update_project(self, project: Project) -> Project:
#         # TODO raise Exception
#         self._save_project(project)
#         return project
#
#     def projects(self) -> List[Project]:
#         return self._list_projects()
#
#     def delete_project(self, name: str):
#         file = self._locate_project(name)
#         file.unlink(True)
#
#     def save_account(self, account: Account) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user: User object to be saved in db
#         """
#         file = self._locate_account(account.name)
#         file.parent.mkdir(parents=True, exist_ok=True)
#
#         # Store user data
#         file.write_text(account.json())
#         # TODO resolving tokens requires a different way of storage
#         # file.write_text(json.dumps({
#         #     "name": account.name
#         # }))
#
#         return account.copy()
#
#     def resolve_token(self, token_id: str) -> Optional[Account]:
#         tokens = self._map_token_to_account()
#         return tokens.get(token_id)
#
#     def load_account(self, account_id: str) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user_id: ID of the user, which will be used to search for
#         """
#         file = self._locate_account(account_id)
#         if file.exists():
#             return Account.parse_file(file)
#
#         return None



"""
Implementation of PEP 503
"""
import logging
import mimetypes
import os
from pathlib import Path
from urllib.parse import urljoin

from flask import Blueprint, request, render_template, redirect, send_file, abort, g
from flask_httpauth import HTTPBasicAuth
from pypitoken import Token, ValidationError, LoaderError
from werkzeug.datastructures import MultiDict

from warehouse14.models import Project, File
from warehouse14.pkg_helpers import normalize_pkgname_for_url
from warehouse14.repos import DBBackend
from warehouse14.storage import PackageStorage

SINGLE_USE_METADATA = {
    "summary",
    "home_page",
    "author",
    "author_email",
    "maintainer",
    "maintainer_email",
    "license",
    "description",
    "keywords",
    "comment",
    "description_content_type",
}

MULTIPLE_USE_METADATA = {
    "project_urls",
    "classifiers",
}

ACCEPTED_METADATA = SINGLE_USE_METADATA | MULTIPLE_USE_METADATA


def extract_metadata(form: MultiDict):
    metadata = {}
    for key, value in form.items():
        if key in ACCEPTED_METADATA:
            if key in MULTIPLE_USE_METADATA:
                metadata[key] = form.getlist(key)
            elif key == "keywords":
                metadata[key] = value.split(",")
            else:
                metadata[key] = value
    return metadata


def create_blueprint(
    db: DBBackend,
    storage: PackageStorage,
    allow_project_creation: bool = False,
    restrict_project_creation=None,
):
    app = Blueprint("simple", __name__)
    token_auth = HTTPBasicAuth()
    log = logging.getLogger(__name__)

    def check_project_creation_allowed(username):
        if not allow_project_creation:
            return False

        return (
            restrict_project_creation is None or username in restrict_project_creation
        )

    @token_auth.error_handler
    def error_handler(status):
        return (
            "Unauthorized: please create an API token using your account page. "
            "Also check that you use '__token__' as username."
        ), status

    @token_auth.verify_password
    def verify_password(username, password):
        """
        Verifies the given api token.

        :param username: name of the account
        :param password: api token with prefix
        :return: username
        """
        try:
            if username and username != "__token__":
                log.warning(f"Access without proper token username {username}")
                return None

            token = Token.load(password)
            account = db.resolve_token(token.identifier)
            if account is None:
                return None

            for tk in db.account_token_list(account.name):
                if tk.id == token.identifier:
                    # delay test to check for specific project access
                    token.check(key=tk.key)

                    # Store token for later
                    g.token = token
                    return account.name
        except (LoaderError, ValidationError):
            log.warning(f"Access with invalid token permitted.")

        # Not authenticated
        return None

    @app.get("/simple/")
    @token_auth.login_required
    def simple_index():
        usern_name = token_auth.current_user()
        links = sorted(
            (p.name, p.normalized_name())
            for p in db.project_list()
            if p.visible(usern_name)
        )
        return render_template("simple/simple.html", links=links)

    @app.route("/simple/<project_name>/")
    @token_auth.login_required
    def simple_packages(project_name):
        # PEP 503: require normalized project
        normalized = normalize_pkgname_for_url(project_name)
        if project_name != normalized:
            log.info(f"Redirect to normalized project name url")
            return redirect(f"/simple/{normalized}/", 301)

        project = db.project_get(project_name)
        if project is None:
            abort(404)

        user_name = token_auth.current_user()
        if not project.visible(user_name):
            abort(401)

        links = (
            (
                os.path.basename(file.filename),
                urljoin(
                    request.path,
                    f"../../packages/{normalized}/{file.filename}#sha256={file.sha256_digest}",
                ),
            )
            for file in project.files
        )
        return render_template("simple/links.html", project=project_name, links=links)

        # packages = sorted(
        #     db.get_project(project),
        #     key=lambda x: (x.parsed_version, x.relfn),
        # )

        # current_uri = request.path
        #
        # links = (
        #     (
        #         os.path.basename(pkg.relfn),
        #         urljoin(current_uri, f"../../packages/{pkg.fname_and_hash}"),
        #     )
        #     for pkg in packages
        # )
        #
        # return render_template("simple/links.html", project=project, links=links)

    @app.route("/packages/<project_name>/<path:filename>")
    @token_auth.login_required
    def server_static(project_name: str, filename: str):
        # PEP 503: require normalized project
        normalized = normalize_pkgname_for_url(project_name)
        if project_name != normalized:
            log.info(f"Redirect to normalized project name url")
            return redirect(f"/simple/{normalized}/{filename}", 301)

        # Check access
        usern_name = token_auth.current_user()
        project = db.project_get(normalized)
        if not project.visible(usern_name):
            abort(401)

        # serve file
        log.info(f"Provide file {filename}")
        file = storage.get(project_name, filename)
        response = send_file(
            file,
            download_name=Path(filename).name,
            mimetype="application/octet-stream",
        )
        return response
        # if config.cache_control:
        #     response.set_header(
        #         "Cache-Control", f"public, max-age={config.cache_control}"
        #     )
        # raise NotFound(f"Not Found ({filename} does not exist)")

    @app.post("/simple/")
    @token_auth.login_required
    def upload():
        form = request.form

        # Get user
        username = token_auth.current_user()
        log.info(f"Upload request from {username}")

        # Validate request, check for required information
        if form[":action"] != "file_upload":
            log.warning(
                f"Wrong action '{form[':action:']}', only supporting 'file_upload'"
            )
            abort(
                403, f"Wrong action '{form[':action:']}', only supporting 'file_upload'"
            )
        if form["protocol_version"] != "1":
            log.warning(f"Wrong protocol_version '{form['protocol_version']}'.")
            abort(403, f"Wrong protocol_version '{form['protocol_version']}'.")
        if "content" not in request.files:
            log.warning(f"No upload content.")
            abort(403, f"No upload content.")

        REQUIRED_METADATA = {"name", "version", "filetype", "summary", "sha256_digest"}
        if not REQUIRED_METADATA.issubset(set(form.keys())):
            log.warning(f"Missing required information.")
            abort(403, f"Missing required information.")

        project_name = form["name"]
        version = form["version"]
        file = request.files["content"]
        sha256_digest = form["sha256_digest"]

        # get or create project
        project = db.project_get(project_name)
        if project is None:
            if not check_project_creation_allowed(username):
                log.warning(f"No permission to create a new project via direct upload.")
                abort(401, f"No permission to create a new project via direct upload.")
            else:
                project = db.project_save(
                    Project(
                        name=project_name, admins=[username], members=[], public=False
                    )
                )

        # Check permissions, only admins are allowed to upload
        if username not in project.admins:
            log.warning(
                f"No permission to upload packages. {username} -> {project.normalized_name()}"
            )
            abort(401, f"No permission to upload packages")

        try:
            # Store file
            file_key = file.filename
            project.add_file(
                version, File(filename=file_key, sha256_digest=sha256_digest)
            )
            storage.add(project.normalized_name(), file_key, file.stream)
        except FileExistsError:
            log.warning("File already exists, overwriting not allowed")
            abort(409, "File already exists, overwriting not allowed")

        # Update metadata
        new_metadata = extract_metadata(form)
        project.versions[version].metadata.update(new_metadata)

        # Update project
        db.project_save(project)
        log.info(f"Uploaded new file for {project.normalized_name()}: {file_key}")

        return {"upload": "successfully"}

    return app



import urllib
from abc import ABCMeta, abstractmethod
from typing import Dict
from urllib.parse import urlencode

import flask_login
from authlib.integrations.base_client.registry import OAUTH_CLIENT_PARAMS
from authlib.integrations.flask_client import OAuth
from flask import Flask, url_for, session, redirect, request, Response
from flask_login import UserMixin

from warehouse14.models import Account


class User(UserMixin):
    def __init__(self, account: Account):
        self.account = account

    def get_id(self):
        return self.account.name

    def __str__(self):
        return self.account.name


class Authenticator(metaclass=ABCMeta):
    @abstractmethod
    def init_app(self, app):
        """
        init flask application, add required callbacks and more
        """
        pass

    @abstractmethod
    def logout(self) -> Response:
        """
        Provide logout response
        """
        pass


class OIDCAuthenticator(Authenticator):
    app: Flask

    def __init__(self, app=None, **kwargs):
        """
        kwargs are passed to the underlying authlib client if they are listed in OAUTH_CLIENT_PARAMS.
        """
        self.oauth = None

        valid_oidc_parameters = OAUTH_CLIENT_PARAMS + ("user_id_field",)
        self.oidc_config = {
            k: v for k, v in kwargs.items() if k.lower() in valid_oidc_parameters
        }

        if app:
            self.init_app(app)

    def init_app(self, app: Flask):
        self.app = app

        self.oauth = OAuth(app)
        self.oauth.register(
            name="oidc",
            client_kwargs={"scope": "openid email profile"},
            **self.oidc_config,
        )

        app.route("/oauth")(self._auth)
        if hasattr(app, "login_manager"):
            # maybe move outside
            app.login_manager.unauthorized_handler(self._redirect_to_auth_server)
        else:
            raise ValueError(
                "The app requires to be initialized with Flask-Login before"
            )

    def logout(self) -> Response:
        url = urllib.parse.urlsplit(request.url)
        logout_url = urllib.parse.urlsplit(
            self.oauth.oidc.load_server_metadata().get("end_session_endpoint")
        )
        logout_url = logout_url._replace(
            query=urlencode({"redirect_uri": f"{url.scheme}://{url.netloc}"})
        )
        return redirect(logout_url.to_url())

    def _redirect_to_auth_server(self):
        # Store origin url to redirect after login
        session["next"] = request.url

        redirect_uri = url_for("_auth", _external=True)
        oidc = self.oauth.oidc
        return oidc.authorize_redirect(redirect_uri)

    def _auth(self):
        oidc_client = self.oauth.oidc
        token = oidc_client.authorize_access_token()
        user: Dict = oidc_client.parse_id_token(token)

        user_id_field = self.app.config.get("OIDC_USER_ID_FIELD", "email")
        self._login_user_with_id(user.get(user_id_field))

        next_url = session.pop("next", "/")
        return redirect(next_url)

    # noinspection PyProtectedMember,PyUnresolvedReferences
    def _login_user_with_id(self, user_id):
        # Instead of passing the Database into the Authenticator, we reuse the setup Flask-Login to load the user
        # While this relies on implementation details, it eases the app setup process
        login_manager: flask_login.LoginManager = self.app.login_manager
        user = login_manager._user_callback(user_id)
        flask_login.login_user(user)
